import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import seaborn as sns

plt.style.use('seaborn')


def plot_timeseries_ndvi_data(ndvi_data=None, output_fig='', title='', time_range=None):

    fig, ax = plt.subplots(figsize=(15, 7))
    sns.lineplot(x=ndvi_data.index,
                 y=ndvi_data['NDVI'],
                 color='navy',
                 linestyle='solid',
                 marker='o',
                 linewidth=1.0,
                 label='Current NDVI')

    sns.lineplot(x=ndvi_data.index,
                 y=ndvi_data['mean'],
                 color='g',
                 label='Historical Mean NDVI',
                 linewidth=2.0)

    ax.fill_between(x=ndvi_data.index,
                    y1=ndvi_data['mean'] - 2.0 * ndvi_data['std'],
                    y2=ndvi_data['mean'] + 2.0 * ndvi_data['std'],
                    alpha=0.1,
                    color='green',
                    label='95% CI')

    ax.set_xlim(time_range[0], time_range[1]) if time_range else None
    ax.xaxis.set_major_locator(mdates.DayLocator(interval=7))
    ax.set_ylim(0.0, 1)
    ax.grid(True)
    ax.set_ylabel('NDVI')
    ax.set_xlabel('Time')
    ax.set_title(title)
    ax.legend()

    plt.xticks(rotation=90)
    fig.autofmt_xdate()

    fig.savefig(output_fig) if output_fig != '' else None
    return fig, ax



from ecoscope.plotting.geomap import GEOMAP
from ecoscope.plotting.plot import SPEEDMAP, timeseries, datacount, seasonal_window, ndvi_seasonal_transition

__all__ = [GEOMAP, SPEEDMAP, timeseries, datacount, seasonal_window, ndvi_seasonal_transition]



import multiprocessing
import os
import typing
from collections import OrderedDict

import geopandas as gp
import pandas as pd
from shapely.geometry import Point

import sys
import logging

logging.basicConfig(
    format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S',
    stream=sys.stdout,
)


def geometry_point(lat_lon):
    if lat_lon and not pd.isna(lat_lon):
        x, y = float(lat_lon.get('longitude')), float(lat_lon.get('latitude'))
        return Point(x, y)
    return Point()


def set_index_column(df, col_name):
    index_column = df.pop(col_name)

    df.insert(0, col_name, index_column)
    return df.set_index(col_name)


def tz_naive(d):
    return d.tzinfo is None or d.tzinfo.utcoffset(d) is None


def export(df, filename='ouput.csv', layer=None, mode='w', **kwargs):
    if isinstance(df, pd.DataFrame) and filename.endswith('.csv'):
        df.to_csv(filename)
    # writing to spatial data.
    elif isinstance(df, gp.GeoDataFrame):
        if filename.endswith('.shp'):
            df.to_file(filename, mode=mode)
        elif filename.endswith('.gpkg'):
            df.to_file(filename, driver='GPKG', layer=layer)
    else:
        df.to_file(filename)


def parallelize_df(df: typing.Iterable, func: typing.Callable, processes: int = 10):
    """
    Using pool of worker processes to offload tasks
    # https://docs.python.org/3/library/multiprocessing.html#using-a-pool-of-workers
    """
    processes = processes or os.cpu_count()
    with multiprocessing.pool.Pool(processes) as p:
        df = pd.concat(p.map(func, df))
    df.reset_index(inplace=True, drop=True)
    return df


def normalize_dict(x, colname=None):
    """normalize dict into individual columns"""
    try:
        _x = x.get(colname)
        index = list(x.index)
        if _x:
            _x = OrderedDict(x[colname])
        else:
            _x = OrderedDict()
        for key in _x.keys():
            if key not in index:
                x[str(key)] = _x.get(key)
            else:
                x['_'.join((colname, str(key)))] = _x.get(key)
        x = x.drop(colname)
    except Exception as e:
        print(str(e))
    finally:
        return x


def flatten_dict(x, colname=None):
    """normalize dict into individual columns"""
    try:
        _y = x.get(colname)
        _x = None
        if (_y is not None) and (_y != ''):
            _x = OrderedDict(x[colname])
        else:
            _x = OrderedDict()
        for key in _x.keys():
            x['_'.join((colname, str(key)))] = _x.get(key)
        x = x.drop(colname)
    except Exception as e:
        print(str(e))
    finally:
        return x


def memorize(func):
    cache = {}

    def wrapper(*args):
        if args in cache:
            return cache[args]
        result = func(*args)
        cache[args] = func(*args)
        return result
    return wrapper



import pandas as pd

import asyncio
import json
import uuid
from datetime import datetime
from functools import partial
from itertools import chain

import numpy as np
import movdata
import pytz
from pydantic import ValidationError, UUID4

from ecoscope.io.client import EarthRangerClient, AsyncEarthRangerClient
from ecoscope.io.schema import Event, Location, Observation
from ecoscope.io.utils import *

logger = logging.getLogger(__name__)
logging.getLogger('backoff').setLevel(logging.ERROR)  # [backoff] Log only when a giveup event occurs.

initialize_er_client = EarthRangerClient


class EarthRangerException(Exception):
    pass


class EarthRanger:
    def __init__(self, er_client: EarthRangerClient):
        """
        Base class for all the earthranger class io.
        provide both async http and synchronous http request.
        """
        self._er_client = er_client
        self.async_erc_lient = AsyncEarthRangerClient(er_client=self._er_client)

    @staticmethod
    def to_geodataframe(df: pd.DataFrame):
        """Converts the pandas dataframe to geodataframe"""
        gdf = gp.GeoDataFrame(df, geometry=df.location.apply(geometry_point), crs=4326)
        gdf.drop(columns=['location'], inplace=True)
        return gdf


class UserIO(EarthRanger):
    def download_users(self):
        """Downloads all the users has permission to view"""
        users = self._er_client.get_users()
        df = pd.DataFrame(users)
        df = df.apply(axis=1, func=normalize_dict, colname='additional_data')
        return df


class EventIO(EarthRanger):

    @memorize
    def _map_eventtype_id(self):
        """Maps event_id to event_value"""
        eventtypes = self._er_client.get_eventtypes()
        map_et_id = {}
        for et in eventtypes:
            eventtype_id, value = et.get('id'), et.get('value')
            map_et_id[value] = eventtype_id
        return map_et_id

    def _get_eventtype_id(self, et_value=None):
        """get the event id given eventtype value"""
        map_et_id = self._map_eventtype_id()
        return map_et_id.get(str(et_value))

    @property
    def event_fieldnames(self):
        return ['id', 'location', 'time',
                'end_time', 'serial_number',
                'message', 'provenance',
                'event_type', 'priority',
                'attributes', 'comment', 'title',
                'reported_by', 'state',
                'related_subjects',
                'eventsource', 'external_event_id',
                'created_at', 'event_details',
                ]

    @staticmethod
    def _coerce_columns_dtype(df, columns: list, dtype):
        if isinstance(columns, list):
            df[columns] = df[columns].astype(dtype)
        return df

    @property
    def columns_of_dtype_list(self):
        return ['is_contained_in', 'files', 'related_subjects', 'patrols', 'patrol_segments', 'updates']

    def _event_details(self, row):
        non_er_columns = set(row.index).difference(self.event_fieldnames)
        non_er_columns.discard('latitude')
        non_er_columns.discard('longitude')
        non_er_columns.discard('reported_by_id')
        non_er_columns.discard('reported_by_content_type')
        non_er_columns.discard('notes')

        event_details = {}
        for column in non_er_columns:
            val = row.get(column)
            if not np.all(pd.isnull(val)):
                event_details[column] = val
        return event_details

    def upload_events(self, df, event_type):
        def create_event(row):
            longitude = row.get('longitude')
            latitude = row.get('latitude')
            content_type = row.get('reported_by_content_type')
            reported_by = row.get('reported_by_id')
            event_details = self._event_details(row)

            try:
                event = Event(event_details=event_details,
                              event_type=event_type,
                              location=Location(longitude=longitude, latitude=latitude),
                              reported_by=dict(id=reported_by,
                                               content_type=content_type), time=row.get('time'))
            except ValidationError as exc:
                logger.error(exc)
                return pd.Series()
            else:
                response = self._er_client.post_event(event)
                if response and row.get('notes'):
                    note_response = self._er_client.post_event_note(event_id=response.get('id'), notes=row.get('notes'))
                    response['notes'] = note_response[0].get('text')
                    response['notes_id'] = note_response[0]['id']
                return pd.Series(response)

        df = df.apply(create_event, axis=1)
        if df.empty:
            return pd.DataFrame()
        else:
            df = self._coerce_columns_dtype(df, self.columns_of_dtype_list, str)
        return self.to_geodataframe(df)

    def download_events(self, event_type: typing.Union[UUID4, str],
                        start: datetime = None,
                        end: datetime = None,
                        page_size: int = 2500):
        """To download events, one can pass eventtype value or eventtype id"""
        try:
            uuid.UUID(event_type)
        except ValueError:
            event_type = self._get_eventtype_id(event_type)

        if isinstance(start, datetime):
            start = start.isoformat()
        if isinstance(end, datetime):
            end = end.isoformat()

        _daterange = json.dumps(dict(date_range={'lower': start or '', 'upper': end or ''}))
        events = asyncio.run(self.async_erc_lient.get_events(event_type=event_type,
                                                             filter=_daterange, page_size=page_size))
        df = pd.DataFrame(chain.from_iterable(events))
        if df.empty:
            return gp.GeoDataFrame(columns=self.event_fieldnames)

        df = df.apply(axis=1, func=normalize_dict, colname='event_details')
        df = df.apply(axis=1, func=normalize_dict, colname='reported_by')
        df = self._coerce_columns_dtype(df, self.columns_of_dtype_list, str)
        return self.to_geodataframe(df)


class SubjectIO(EarthRanger):

    def download_subjects(self,
                          subjectgroup_name: str = None,
                          include_inactive: bool = False,
                          include_hidden: bool = True,
                          isvisible: bool = True,
                          flat: bool = True
                          ) -> pd.DataFrame:
        """Download subject in a group"""

        def to_hex(val):
            if val and not pd.isnull(val):
                return '#{:02X}{:02X}{:02X}'.format(*[int(i) for i in val.split(',')])
            return np.NaN

        if subjectgroup_name is None:
            subjects = self._er_client.get_subjects(include_inactive=include_inactive)
            df = pd.DataFrame(subjects)
        else:
            subjectgroups = self._er_client.get_subjectgroups(group_name=subjectgroup_name,
                                                              include_inactive=include_inactive,
                                                              include_hidden=include_hidden,
                                                              isvisible=isvisible,
                                                              flat=flat)
            arr = []
            for sg in subjectgroups:
                sg_df = pd.DataFrame([s for s in sg.get('subjects')])
                sg_df['subjectgroup_name'] = sg.get('name')
                arr.append(sg_df)
            df = pd.concat(arr)

        df = df.rename(columns={'id': 'subject_id', 'name': 'subject_name'})
        df = df.drop('content_type', axis=1)
        df.set_index('subject_id', inplace=True)
        df = df.apply(axis=1, func=normalize_dict, colname='additional')
        df['hex'] = df['rgb'].map(to_hex)
        return df

    def download_subject_sources(self, subject_id: UUID4):
        subject_source = self._er_client.get_subject_sources(subject_id)
        return pd.DataFrame(subject_source)

    def download_subject_observations(self,
                                      subject_id: UUID4,
                                      start: datetime = None,
                                      end: datetime = None,
                                      filter_flag: int = 0,
                                      include_details: bool = True,
                                      page_size: int = 2500):
        observations = self._er_client.get_subject_observations(subject_id=subject_id,
                                                                start=start,
                                                                end=end,
                                                                filter_flag=filter_flag,
                                                                include_details=include_details,
                                                                page_size=page_size)
        return pd.DataFrame(observations)


class SourceIO(EarthRanger):

    def download_sources(self, sources=None):
        if sources is not None:
            sources = asyncio.run(self.async_erc_lient.get_sources(sources))
            df = pd.DataFrame(chain.from_iterable(sources))
        else:
            sources = self._er_client._get('sources')
            df = pd.DataFrame(sources.get('results'))

        df = df.rename(columns={'id': 'source_id'})
        df.set_index('source_id', inplace=True)
        return df

    def download_subjectsources(self,
                                subjects: typing.List[str] = None,
                                sources: typing.List[str] = None,
                                page_size: int = 250) -> pd.DataFrame:
        """
        Download SubjectSourceAssignment.

        Parameters
        ----------
        subjects: list[str], optional
            List of subjects ids (UUID)
        sources: list[str], optional
            List of sources ids (UUID)
        page_size: int, default 250
            Integer representing number of results from HTTP request.
        """

        if sources is None:
            sources = []
        if subjects is None:
            subjects = []

        subjects = ','.join(subjects)
        sources = ','.join(sources)

        subject_sources = asyncio.run(self.async_erc_lient.get_subjectsources(subjects=subjects,
                                                                              sources=sources,
                                                                              page_size=page_size))
        df = pd.DataFrame(chain.from_iterable(subject_sources))
        df = df.apply(axis=1, func=normalize_dict, colname='assigned_range')

        if not df.empty:
            df['subject_source_start'] = pd.to_datetime(df['lower'], errors='coerce')
            df['subject_source_end'] = pd.to_datetime(df['upper'], errors='coerce')
            df = df.rename(columns={'id': 'subject_source_id',
                                    'source': 'source_id',
                                    'subject': 'subject_id',
                                    'additional': 'subject_source_additional'})
            df.set_index('subject_source_id', inplace=True)

            # safely drop unwanted columns
            un = df.columns[df.columns.str.startswith('assigned_range_')]
            df.drop(un, axis=1, inplace=True)
        return df


class ObservationsIO(EarthRanger):

    # @exception_handler
    # def _post_observation(self, payload, sensor_type):
    #     self._er_client.post_sensor_observation(observation=payload, sensor_type=sensor_type)

    @property
    def observation_fieldnames(self):
        return ['id', 'location', 'recorded_at', 'additional', 'manufacturer_id']

    def _additional_data(self, row):
        non_er_columns = set(row.index).difference(self.observation_fieldnames)
        non_er_columns.discard('X')
        non_er_columns.discard('Y')

        obs_additional = {}
        for column in non_er_columns:
            val = row.get(column)
            if not np.all(pd.isnull(val)):
                obs_additional[column] = val

        obs_additional['externalguid'] = str(row.name)
        return obs_additional

    def upload_observations(self, df=None, sensor_type='generic', source_type='tracking-device'):
        def create_observation(row):
            longitude = row.get('X')
            latitude = row.get('Y')
            recorded_at = row.get('recorded_at')
            manufacturer_id = row.get('manufacturer_id')
            additiona = self._additional_data(row)

            try:
                observation = Observation(manufacturer_id=manufacturer_id,
                                          source_type=source_type,
                                          recorded_at=recorded_at,
                                          additional=additiona,
                                          location=Location(x=longitude, y=latitude)
                                          )

            except ValidationError as exc:
                logger.exception(exc)
            else:
                self._er_client.post_sensor_observation(observation=observation, sensor_type=sensor_type)
                return pd.Series(observation)

        return df.apply(create_observation, axis=1)

    def _get_subject_observations(self,
                                  subject,
                                  start, end,
                                  filter_flag,
                                  include_additional,
                                  include_subject_details,
                                  chunk_size,
                                  limit,
                                  timezone,
                                  count=None,
                                  ):
        try:
            logger.info(f"Querying observations for subject: {subject.get('subject_name')}")

            result = asyncio.run(self.async_erc_lient.get_subject_observations(
                subject_id=subject['subject_id'],
                start=start,
                end=end,
                filter_flag=filter_flag,
                include_details=include_additional,
                page_size=chunk_size,
                count=count,
                limit=limit
            ))
            result = chain.from_iterable(result)
            df = pd.DataFrame(result)

            if df.empty:
                return df

            df = df.rename(columns={'source': 'source_id', 'id': 'observation_id', 'recorded_at': 'fixtime'})

            # Parse the fixtime and created_at columns as datetime
            df['fixtime'] = pd.to_datetime(df['fixtime'], utc=timezone)
            df['created_at'] = pd.to_datetime(df['created_at'], utc=timezone)
            df = df.astype({'observation_details': str}, errors='ignore')

            # default subject column in observation df
            df['subject_id'] = subject.get('subject_id')

            if include_subject_details:
                # Add subject related info
                df['subject_name'] = subject.get('subject_name')
                df['subject_sex'] = subject.get('sex')
                df['subject_type'] = subject.get('subject_type')
                df['subject_subtype'] = subject.get('subject_subtype')
                df['region'] = subject.get('region')
                df['country'] = subject.get('country')

                primary_cols = ['observation_id', 'source_id', 'subject_id',
                                'subject_name', 'subject_type', 'subject_subtype',
                                'subject_sex', 'region', 'country', 'fixtime',
                                'created_at', 'location']
            else:
                primary_cols = ['observation_id', 'source_id', 'subject_id', 'fixtime', 'created_at', 'location']

            if include_additional:
                df = df.rename(columns={'observation_details': 'additional'})

            # re-order columns
            other_cols = [col for col in df.columns if col not in primary_cols]
            return df[primary_cols + other_cols]
        except AttributeError as exc:
            logger.exception(exc)

    def download_observations(self,
                              subjectgroup_name='',
                              include_inactive=True,
                              start=None,
                              end=None,
                              filter_flag=0,
                              include_additional=True,
                              include_subject_details=False,
                              include_source_details=False,
                              include_subject_source_details=False,
                              chunk_size=500,
                              timezone=pytz.UTC,
                              limit=100,
                              processes=None
                              ):
        """
             Passing None, returns everything
             Passing 0, filters out everything but rows with exclusion flag 0 (i.e, clean data)
             Passing 1, filters out everything but rows with exclusion flag 1 (i.e., manually filtered data)
             Passing 2, filters out everything but rows with exclusion flag 2 (i.e., automatically filtered data)
             Passing 3, filters out everything but rows with exclusion flag 2 or 1 (i.e., both manual and automatic filtered data)

             Maximum page_size per request is 4000 according to observation api.

             Arg: limit - limit amount of simultaneously opened connections; default to 100 &
             If you explicitly want not to have limits, pass 0.

             Arg: processes - number of worker processes to use.
             If processes is None then the number returned by os.cpu_count() is used.
        """
        subject_io = SubjectIO(self._er_client)
        subjects = subject_io.download_subjects(subjectgroup_name=subjectgroup_name, include_inactive=include_inactive)

        if subjects is None:
            raise EarthRangerException(f'The subjects dataframe for {subjectgroup_name} is null')

        # Assume UTC if no timezone is given
        if start is not None and tz_naive(start):
            start = start.replace(tzinfo=pytz.UTC)
        if end is not None and tz_naive(end):
            end = end.replace(tzinfo=pytz.UTC)

        frames = [sub for _, sub in subjects.reset_index().iterrows()]

        with multiprocessing.Manager() as manager:
            # Manager controls a server process which holds Python objects allows other processes to manipulate them.
            count = manager.list()
            obs_df = parallelize_df(frames, func=partial(self._get_subject_observations,
                                                         start=start, end=end,
                                                         filter_flag=filter_flag,
                                                         include_additional=include_additional,
                                                         include_subject_details=include_subject_details,
                                                         chunk_size=chunk_size,
                                                         count=count,
                                                         timezone=timezone,
                                                         limit=limit
                                                         ), processes=processes)

            count = sum(count)
            if count != obs_df.shape[0]:
                logger.warning(f"Partial Success: Only {obs_df.shape[0]} out of {count} were successfully downloaded.")

        if obs_df.empty:
            return gp.GeoDataFrame()

        obs_df.sort_values(by=['fixtime'], axis=0)

        if include_subject_source_details:
            source_io = SourceIO(self._er_client)
            subject_source_df = source_io.download_subjectsources(sources=obs_df.source_id.unique()).reset_index()
            subject_source_df = subject_source_df[['subject_source_id',
                                                   'source_id',
                                                   'subject_source_start',
                                                   'subject_source_end',
                                                   'subject_source_additional']]

            obs_df = pd.merge(subject_source_df, obs_df, on='source_id')

        if include_source_details:
            source_io = SourceIO(self._er_client)
            source_df = source_io.download_sources(obs_df.source_id.unique())
            source_df = source_df[['manufacturer_id',
                                   'model_name',
                                   'provider',
                                   'source_type']]
            obs_df = pd.merge(source_df, obs_df, left_index=True, right_on='source_id')

        obs_df = set_index_column(obs_df, 'observation_id')
        return self.to_geodataframe(obs_df)

    @staticmethod
    def multi_relocations(obs_df, filter_point_coords=None, max_speed_kmhr=7.0):
        if filter_point_coords is None:
            filter_point_coords = [[180, 90]]

        multi_relocs = movdata.base.MultiRelocations.from_observations(obs_df)

        junk_pnts_filter = movdata.base.RelocsCoordinateFilter(filter_point_coords=filter_point_coords)
        multi_relocs.apply_fix_filter(junk_pnts_filter)

        # and let's apply a speed filter to get rid of any point where the animal had to move faster than X to get
        # there from the previous position
        speed_filter = movdata.base.RelocsSpeedFilter(max_speed_kmhr=max_speed_kmhr)
        multi_relocs.apply_fix_filter(speed_filter)

        return multi_relocs

    def multi_trajectory(self, obs_df, filter_point_coords=None, max_speed_kmhr=8.0, max_time_secs=14400):
        multi_relocs = self.multi_relocations(obs_df, filter_point_coords, max_speed_kmhr)

        multi_trajs = movdata.base.MultiTrajectory(multi_relocs)
        traj_seg_filter = movdata.base.TrajSegFilter(max_speed_kmhr=max_speed_kmhr, max_time_secs=max_time_secs)
        multi_trajs = multi_trajs.apply_traj_filter(traj_seg_filter)
        return multi_trajs

    def relocation(self, obs_df, subject_id, filter_point_coords=None, max_speed_kmhr=7.0):
        obs_df = obs_df.loc[obs_df.subject_id == str(subject_id)]
        return self.multi_relocations(obs_df, filter_point_coords, max_speed_kmhr)

    def trajectory(self, obs_df, subject_id, filter_point_coords=None, max_speed_kmhr=7.0, max_time_secs=14400):
        obs_df = obs_df.loc[obs_df.subject_id == str(subject_id)]
        return self.multi_trajectory(obs_df, filter_point_coords, max_speed_kmhr, max_time_secs)



import os
import typing

import matplotlib.pyplot as plt
from geopy.distance import great_circle
from matplotlib_scalebar.scalebar import ScaleBar

from ecoscope.plotting._accesory import plot_vector_layer, plot_raster_layer, plot_tiled_layer
from ecoscope.plotting.struct import LAYERS

plt.style.use('seaborn')


class GEOMIXIN:

    @staticmethod
    def add_scalebar(ax, **kwargs):
        minx, maxx, miny, maxy = ax.axis()
        midpoint = (miny + maxy) / 2
        p1, p2 = (midpoint, int(minx)), (midpoint, int(minx) + 1)
        dist = great_circle(p1, p2).m
        ax.add_artist(ScaleBar(dist, location=kwargs.get("scalebar_location", "lower right")))

    @staticmethod
    def add_north_arrow(ax):
        x, y, arrow_length = 0.05, 0.98, 0.04
        ax.annotate('N', xy=(x, y), xytext=(x, y - arrow_length),
                    arrowprops=dict(facecolor='black', width=5, headwidth=15),
                    ha='center', va='center', fontsize=20,
                    xycoords=ax.transAxes)

    @staticmethod
    def set_extent(extent, ax):
        # extent:  MaxX, MaxY, MinX, and MinY.
        xlim = ([extent[0], extent[2]])
        ylim = ([extent[1], extent[3]])
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

    @staticmethod
    def apply_extent_ratio(extent_ratio, ax):
        # minx, miny, maxx, maxy = total_bounds

        minx, maxx, miny, maxy = ax.axis()
        x = maxx - minx
        y = maxy - miny

        # x-dimension
        new_x = x * extent_ratio
        diff_x = new_x - x
        minx = minx - 0.5 * diff_x
        maxx = maxx + 0.5 * diff_x

        # y-dimension:
        new_y = y * extent_ratio
        diff_y = new_y - y
        miny = miny - 0.5 * diff_y
        maxy = maxy + 0.5 * diff_y

        xlim = ([minx, maxx])
        ylim = ([miny, maxy])

        ax.set_xlim(xlim)
        ax.set_ylim(ylim)


class GEOMAPLAYERS(LAYERS):
    @property
    def layers(self):
        return self

    @property
    def add_vector_layer(self):
        return self.vector_layer

    @property
    def add_tiled_layer(self):
        return self.tiled_layer

    @property
    def add_raster_layer(self):
        return self.raster_layer


class GEOMAP(GEOMAPLAYERS, GEOMIXIN):
    def __init__(self,
                 title: typing.Optional[str] = '',
                 figsize: typing.Tuple[int, int] = (15, 15),
                 legend: bool = True,
                 north_arrow: bool = True,
                 scale_bar: bool = True,
                 extent_ratio: float = 1.,
                 extent: typing.List[float] = None,
                 crs: typing.Any = 4326
                 ):
        self.figure, self.axis = plt.subplots(1, figsize=figsize)
        self._title = title
        self._figsize = figsize
        self._legend = legend
        self._north_arrow = north_arrow
        self._scalebar = scale_bar
        self._extent = extent or []
        self._extent_ratio = extent_ratio
        self._crs = crs
        super(GEOMAP, self).__init__(title=title,
                                     figsize=figsize,
                                     legend=legend,
                                     north_arrow=north_arrow,
                                     scale_bar=scale_bar,
                                     extent_ratio=extent_ratio,
                                     extent=extent)
        if len(list(self._extent)) == 4:
            self.set_extent(self._extent, self.axis)
        self.axis.set_aspect('equal', adjustable='datalim', anchor='C')

    def _add_geoplot_elements(self, **kwargs: typing.Any):
        frameon = kwargs.get('frameon', True)
        legend_prop = {'size': kwargs.get('legend_fontsize', 18)}
        legend_location = kwargs.get('legend_location', 'best')
        title_fontdict = {'fontsize': kwargs.get('title_fontsize', 25),
                          'fontweight': kwargs.get('title_fontweight', 'medium')}

        if self._legend:
            self.axis.legend(prop=legend_prop, loc=legend_location, frameon=frameon)
        if self._north_arrow:
            self.add_north_arrow(self.axis)
        if self._scalebar:
            self.add_scalebar(self.axis, **kwargs)
        if self._title:
            self.axis.set_title(self._title, fontdict=title_fontdict)

    def _reset_extent_ratio(self):
        # Temporary-Fix: Render the figure in order to get the the auto-scaled x and y dimension.
        # This will be used to create basemap that will align with axis box.
        try:
            self.figure.canvas.print_figure('_xufgremdfdefs')
            os.remove('_xufgremdfdefs.png')
        except OSError:
            pass
        self.apply_extent_ratio(self._extent_ratio or 1.0, self.axis)

    def plot(self, *args, **kwargs):
        layers = self.layers.keys(multi=True)

        for index, layer in reversed(list(enumerate(layers, 1))):
            layer_props = self.layers[layer]
            plot_kwargs = layer_props.setdefault('properties', {})

            if 'geo_dataframe' in layer:
                plot_vector_layer(self.axis,
                                  dataframe=layer_props['df'],
                                  zorder=index, **plot_kwargs)
            if 'raster_layer' in layer:
                plot_raster_layer(self.axis,
                                  raster_path=layer_props['raster'],
                                  resampling=layer_props['resampling'],
                                  crs=layer_props['crs'],
                                  zorder=index, **plot_kwargs)
            if 'tiled_layer' in layer:
                self._reset_extent_ratio()
                plot_tiled_layer(axis=self.axis,
                                 source=layer_props['source'],
                                 crs=plot_kwargs.pop('crs', self._crs),
                                 zorder=index, **plot_kwargs)

        self._add_geoplot_elements(**kwargs)
        return self.axis, self.figure

    def savefig(self, fname, **kwargs):
        self.figure.savefig(fname=fname, **kwargs)



import abc
import logging
from enum import Enum
from typing import List, Optional, Dict, Any, Iterable
from typing import TypeVar
from typing import Union
from uuid import UUID
from datetime import datetime, timezone

from pydantic import BaseModel, Field, HttpUrl, ValidationError, validator, UUID4, Json


class Location(BaseModel):
    longitude: float = Field(..., ge=-180.0, le=360.0, title="Longitude in decimal degrees")
    latitude: float = Field(..., ge=-90.0, le=90.0, title="Latitude in decimal degrees")


class Event(BaseModel):
    id: Optional[UUID4] = Field(None, title='Event ID', description='The unique ID for the event')
    serial_number: Optional[int] = Field(None, title='Serial Number')
    message: Optional[Any] = Field(None, title='Event Message')
    comment: Optional[str] = Field(None, description='Additional message text')
    title: Optional[str] = Field(None, title='Event Title')
    time: datetime = Field(..., title='Event Time', description='Timestamp for the data, preferrably in ISO format.',
                           example='2021-03-21 12:01:02-0700')
    end_time: Optional[datetime]
    provenance: Optional[str]
    priority: Optional[int]
    state: Optional[str]
    event_type: str
    location: Location
    attributes: Optional[Json]
    created_at: Optional[datetime]
    # reported_by: Optional[dict]
    event_details: Optional[Dict[str, Any]]

    class Config:
        title = 'Event'


class Observation(BaseModel):
    id: Optional[UUID4]
    manufacturer_id: str
    source_type: Optional[str]
    recorded_at: datetime
    location: Location
    additional: Optional[Dict[str, Any]]


import copy
import warnings

import contextily
import numpy
import rasterio

from ecoscope.plotting.geo_util import warp


class PLOTACCESSORY:
    @classmethod
    def plot_vector_layer(cls, axis, dataframe, zorder, **plot_kwargs):
        column = plot_kwargs.pop('column', None)
        if column:
            # Todo: leverage on geopandas plot_dataframe method
            df = dataframe[[column, 'geometry']]
            df.set_index(column, inplace=True)

            _kwargs = copy.copy(plot_kwargs)
            labels = dict(zip(df.index.unique(), _kwargs.setdefault('label', [])))
            colors = dict(zip(df.index.unique(), _kwargs.setdefault('color', [])))

            if labels or colors:
                plot_kwargs.pop('label', None)
                plot_kwargs.pop('color', None)
                plot_kwargs.pop('column', None)
                return [geom.plot(ax=axis,
                                  color=colors.get(value),
                                  label=labels.get(value, ''),
                                  zorder=zorder, **plot_kwargs)
                        for value, geom in df.groupby(column) if not geom.is_empty.all()]
        return dataframe.plot(ax=axis, column=column, zorder=zorder, **plot_kwargs)

    @classmethod
    def plot_raster_layer(cls, axis, raster_path, zorder, crs, resampling, **raster_kwargs):
        with rasterio.open(raster_path) as src:
            dataset = src.read()
            transform = src.transform
            dataset[dataset == src.nodata] = numpy.nan
            bb = src.bounds
            extent = bb.left, bb.right, bb.bottom, bb.top

            # Warp
            if src.crs != crs and crs is not None:
                dataset, bounds, _ = warp(dataset, transform, src.crs, crs, resampling)
                extent = bounds.left, bounds.right, bounds.bottom, bounds.top
            dataset = dataset.transpose(1, 2, 0)

        # plotting
        if dataset.shape[2] == 1:
            dataset = dataset[:, :, 0]
        cmap = raster_kwargs.pop('cmap', 'RdYlGn_r')
        return axis.imshow(dataset, extent=extent, cmap=cmap, zorder=zorder, **raster_kwargs)

    @classmethod
    def plot_tiled_layer(cls, axis, source, crs, zorder, **cx_kwargs):
        try:
            return contextily.add_basemap(ax=axis, source=source, crs=crs, zorder=zorder, **cx_kwargs)
        except Exception as exc:
            warnings.warn(str(exc))


plot_vector_layer = PLOTACCESSORY.plot_vector_layer
plot_raster_layer = PLOTACCESSORY.plot_raster_layer
plot_tiled_layer = PLOTACCESSORY.plot_tiled_layer



import os
from setuptools import setup
import re

version_file = open("ecoscope/version.py").read()
version_data = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", version_file))

try:
    descr = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()
except IOError:
    descr = ""

setup(
    name='ecoscope',
    version=version_data['version'],
    description='Standard Analytical Reporting Framework for Conservation',
    long_description=descr,
    url='http://github.com/walljcg/conservation-macroscope',
    author='Jake Wall',
    author_email='walljcg@gmail.com',
    license='MIT',
    packages=[
        "ecoscope",
        "ecoscope.io",
        "ecoscope.plotting",
        "ecoscope.analysis"
    ],
    platforms=u'Posix; MacOS X; Windows',
    install_requires=['pandas',  'xlrd', 'seaborn',
                      'openpyxl', 'pydantic', 'aiohttp', 'backoff', 'boltons', 'shapely<=1.7.1',
                      'contextily', 'matplotlib_scalebar', 'mapclassify', 'fsspec>=0.3.3', 'movdata'],  # 'dasclient'
    zip_safe=False,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

    ]
)



import typing

import geopandas
import mapclassify
import pandas
import itertools
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import seaborn as sns
from shapely.geometry import MultiLineString
from ecoscope.plotting import GEOMAP

ESRI_TOPOLOGY_BASEMAP = 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer/tile/{z}/{y}/{x}'


class SPEEDMAP(GEOMAP):
    speedmap_labels = []

    @property
    def default_speed_color(self):
        return ['#1a9850', '#91cf60', '#d9ef8b', '#fee08b', '#fc8d59', '#d73027']

    @staticmethod
    def _speedmap_labels(bins):
        return [f'{bins[i]:.1f} - {bins[i + 1]:.1f} km/hr' for i in range(len(bins) - 1)]

    @property
    def classification_methods(self):
        return {
            "equal_interval": mapclassify.EqualInterval,
            "natural_breaks": mapclassify.NaturalBreaks,
            "quantile": mapclassify.Quantiles,
            "std_mean": mapclassify.StdMean,
            "max_breaks": mapclassify.MaximumBreaks,
            "fisher_jenks": mapclassify.FisherJenks,
        }

    @staticmethod
    def create_multi_linestring(s):
        return MultiLineString(s['geometry'].tolist())

    def apply_classification(self, x, k, cls_method='natural_breaks', multiples=None):
        """
        Function to select which classifier to apply to the speed distributed data.

        Args:
        __________
        x (array)          : The input array to be classified. Must be 1-dimensional
        k (int)            : Number of classes required.
        cls_method (str)   : Classification method
        multiples (array)  : the multiples of the standard deviation to add/subtract from
                             the sample mean to define the bins.
                             defaults=[-2,-1,1,2]
        """
        if multiples is None:
            multiples = [-2, -1, 1, 2]

        classifier = self.classification_methods.get(cls_method)
        if not classifier:
            return

        map_classifier = classifier(x, multiples) if cls_method == 'std_mean' else classifier(x, k)
        edges, _, _ = mapclassify.classifiers._format_intervals(map_classifier, fmt="{:.2f}")

        print(map_classifier)
        return [float(i) for i in edges]

    def create_speedmap_df(self,
                           multi_trajectory: geopandas.GeoDataFrame,
                           classification_method: str = 'equal_interval',
                           no_class: int = 6,
                           speed_colors: typing.List = None
                           ):
        if speed_colors is None:
            speed_colors = self.default_speed_color[:no_class]

        # apply classification on speed data.
        bins = self.apply_classification(multi_trajectory.speed_kmhr, no_class, cls_method=classification_method)
        multi_trajectory['speed_colour'] = pandas.cut(x=multi_trajectory.speed_kmhr,
                                                      bins=bins, labels=speed_colors,
                                                      include_lowest=True)
        # Group the data according to speed and create multi-linestrings for each group
        speedmap_df = geopandas.GeoDataFrame(geometry=multi_trajectory.groupby('speed_colour').apply(
            self.create_multi_linestring), crs=4326).reset_index()

        speedmap_df.sort_values(by='speed_colour', inplace=True)
        self.speedmap_labels = self._speedmap_labels(bins)
        return speedmap_df

    def _add_layers(self, df, basemap, linewidth):
        self.add_tiled_layer(source=basemap, crs=df.crs)
        self.add_vector_layer(df,
                              color=df.speed_colour.unique(),
                              column='speed_colour',
                              linewidth=linewidth,
                              linestyle='-', marker=0,
                              label=self.speedmap_labels)

    def plot_speedmap(self,
                      multi_trajectory: geopandas.GeoDataFrame,
                      basemap=ESRI_TOPOLOGY_BASEMAP,
                      classification_method: str = 'equal_interval',
                      linewidth: float = 1.,
                      no_class: int = 6,
                      speed_colors: typing.List = None,
                      **kwargs: typing.Any
                      ):

        df = self.create_speedmap_df(multi_trajectory=multi_trajectory,
                                     classification_method=classification_method,
                                     no_class=no_class,
                                     speed_colors=speed_colors)
        self._add_layers(df=df, basemap=basemap, linewidth=linewidth)
        return super(SPEEDMAP, self).plot(**kwargs)


# alias
SpeedMap = SPEEDMAP


def timeseries(df,
               x='x',
               y='y',
               xlabel='',
               ylabel='y',
               title='',
               legend_label='',
               fmt='g',
               figsize=(15, 7),
               linestyle='solid'):
    fig, ax = plt.subplots(figsize=figsize)
    ax.xaxis_date()
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
    ax.grid(True)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)
    ax.set_title(title)
    ax.plot_date(df[x], df[y], fmt=fmt, label=legend_label, linewidth=1.5, linestyle=linestyle)
    fig.autofmt_xdate()
    ax.legend(loc=0)
    return fig, ax


def datacount(data, grouping_column, figsize=(20, 10), title='', xlabel=None, ylabel=None, rotation=90, **countplot):
    sns.set(rc={'figure.figsize': figsize})
    ax = sns.countplot(x=grouping_column, data=data, **countplot)
    plt.xticks(rotation=rotation)
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    return ax


def seasonal_window(data,
                    x='start',
                    y='season_code',
                    xlabel='',
                    ylabel='Season (1=wet, 0=dry)',
                    draw_style='steps-post',
                    step='post',
                    color='blue',
                    alpha=0.3):
    ax = sns.lineplot(x=x, y=y, data=data, drawstyle=draw_style)
    l1 = ax.lines[0]
    x1 = l1.get_xydata()[:, 0]
    y1 = l1.get_xydata()[:, 1]
    ax.fill_between(x1, y1, step=step, color=color, alpha=alpha)
    ax.set(xlabel=xlabel, ylabel=ylabel)
    return ax


def ndvi_seasonal_transition(vals,
                             season_cut_values,
                             xlabel='',
                             ylabel='',
                             title='',
                             subplot_kwargs=None,
                             histplot_kwargs=None,
                             axvline_kwargs=None):
    xmin = min(vals) - 0.1 * min(vals)
    xmax = max(vals) + 0.1 * min(vals)
    fig, ax = plt.subplots(**subplot_kwargs)
    ax.set_xlim(xmin, xmax)
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    sns.histplot(vals, **histplot_kwargs, ax=ax)
    [ax.axvline(x=i, **axvline_kwargs) for i in season_cut_values[1:-1]]
    return ax
